/*****************************************************************************
 *              D: a minimal compiler for educational purpose                *
 ***************************************************************************** 
 *                                                                           *
 * Copyright (c) 2009 Christophe Alias                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the     *
 * Free Software Foundation; either version 2 of the License, or (at your    * 
 * option) any later version.                                                *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General  *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * Written by Christophe Alias, Christophe.Alias@ens-lyon.fr                 *
 *                                                                           *
 *****************************************************************************/
/* ============================================================== */
/*                             C PRELUDE                          */
/* ============================================================== */

%{
#include <stdlib.h>
#include <stdio.h>
#include "label.h"
#include "register.h"
#include "symbol_table.h"
#include "code_generation.h"   

int		line_number	= 1;	/* number of current source line */
extern int	yylex();	/* lexical analyzer generated from lex.l */
extern char	*yytext;	/* last token, defined in lex.l  */

void yyerror(char *s)
  { fprintf(stderr, 
	    "d parser:%d: syntax error. Last token was \"%s\"\n",
	    line_number, yytext);
    exit(1);
  }
void error(char *s)
  { fprintf(stderr, 
	    "d parser:%d: %s\n", line_number, s);
    exit(1);
  }
%}

/* ============================================================== */
/*                     TYPES OF NON-TERMINALS                     */         
/* ============================================================== */

%union
{
  int integer;
  char* id_string;
}


/* ============================================================== */
/*                       TOKENS DECLARATION
/* ============================================================== */

/* dlimiteurs */
%token TK_LPAR TK_RPAR
%token TK_COMMA TK_SEMI

/* oprations boolennes */
%token TK_TRUE TK_FALSE

/* oprations mixtes */
%token TK_GREATER TK_EQUAL TK_NEQUAL

/* oprations arithmtiques */
%token TK_MUL TK_DIV TK_PLUS TK_MINUS
%token TK_NUM
%token TK_ID

/* dclarations */
%token TK_INT

/* controle */
%token TK_ASSIGN TK_IF TK_THEN TK_ELSE TK_FI TK_WHILE TK_DO TK_DONE
%token TK_RETURN

/* fonctions */
%token TK_FUNCTION TK_BEGIN TK_END


/* associativit */
%left TK_PLUS TK_MINUS
%left TK_MUL TK_DIV

%type<integer> TK_NUM
%type<id_string> TK_ID

%type<id_string> e_expr
%type<id_string> t_expr
%type<id_string> f_expr
%type<id_string> expr
%type<id_string> test

/* axiom */
%start prog

%%

/* ============================================================== */
/*               GRAMMAR ITSELF WITH SEMANTIC ACTIONS             */
/* ============================================================== */

/* -------------------------------------------------------------- */
/*                  1/ Expressions (e.g. 2*x+y)                   */
/* -------------------------------------------------------------- */

//ETF (sub-)grammar

e_expr: 
  e_expr TK_PLUS t_expr 
    { 
      char* result_register = new_register(); 
      add(result_register,$1,$3);
      $$ = result_register;
    }
| e_expr TK_MINUS t_expr 
    { 
      char* result_register = new_register(); 
      sub(result_register,$1,$3);
      $$ = result_register;
    }
| t_expr 
    { $$ = $1; }
;

t_expr: 
  t_expr TK_MUL f_expr
    { 
      char* result_register = new_register(); 
      mul(result_register,$1,$3);
      $$ = result_register;
    }
| f_expr
    { $$ = $1; }
;

f_expr: 
  TK_NUM
    { 
      char* result_register = new_register(); 
      move_int(result_register,$1);
      $$ = result_register;
    }
| TK_ID 
    { $$ = get_register($1); }
| TK_ID TK_LPAR caller_arg_list TK_RPAR
    { call($1); $$ = RETURN_REGISTER; }
| TK_LPAR e_expr TK_RPAR 
    { $$ = $2; }
;

expr: 
  e_expr
    { $$ = $1; }
;

/* -------------------------------------------------------------- */
/*                2/ Conditions (e.g. 2*x+y != 0)                 */
/* -------------------------------------------------------------- */

test: 
  expr TK_GREATER expr 
    { 
      char* result_register = new_register(); 
      cmp_g(result_register,$1,$3);
      $$ = result_register;
    }
| expr TK_EQUAL expr
    { 
      char* result_register = new_register(); 
      cmp_eq(result_register,$1,$3);
      $$ = result_register;
    }
| expr TK_NEQUAL expr
    { 
      char* result_register = new_register(); 
      cmp_neq(result_register,$1,$3);
      $$ = result_register;
    }
| TK_LPAR test TK_RPAR
    { $$ = $2; }
;

/* -------------------------------------------------------------- */
/*                 3/ Statements (e.g. if, while)                 */
/* -------------------------------------------------------------- */


stmt:

/* ---                ASSIGNMENT (e.g. x := e;)                -- */

  TK_ID TK_ASSIGN expr                             
    { move_reg(get_register($1),$3); }


/* ---                    CONDITIONAL (IF)                     -- */

| TK_IF test TK_THEN 
    { 
      char* label_else = new_label();
      jmp_if_false($2,label_else);
      push_label(label_else);
    }
  stmt_list TK_ELSE
    { 
      char* label_endif = new_label();
      jmp_lbl(label_endif);
      char* label_else = pop_label();
      lbl(label_else); //put the label
      push_label(label_endif);
    }
  stmt_list
  TK_FI
    {
      char* label_endif = pop_label();
      lbl(label_endif);
    }

/* ---                       WHILE                             -- */

| TK_WHILE
    {
      char* label_next = new_label();
      lbl(label_next);
      push_label(label_next);
    } 
  test TK_DO
    {
      char* label_end = new_label();
      jmp_if_false($3,label_end);
      push_label(label_end);
    }
  stmt_list TK_DONE
    { 
      char* label_end = pop_label();
      char* label_next = pop_label();
      jmp_lbl(label_next);
      lbl(label_end);
    }

/* ---                       RETURN                            -- */

| TK_RETURN expr
    { move_reg(RETURN_REGISTER,$2); }
;

stmt_list:
| stmt_list stmt
;

/* -------------------------------------------------------------- */
/*                 4/ Declarations (e.g. int x,y;)                */
/* -------------------------------------------------------------- */

arg_list:
  TK_ID
    { reset_symbols(); add_argument($1); }
| arg_list TK_COMMA TK_ID
    { add_argument($3); }
;

decl_args: 
| TK_INT arg_list
;

local_var_list:
  TK_ID
    { add_local_var($1); }
| local_var_list TK_COMMA TK_ID
    { add_local_var($3); }
;

decl_local_vars: 
| TK_INT local_var_list
;

/* -------------------------------------------------------------- */
/*                 5/ Functions                                   */
/* -------------------------------------------------------------- */

caller_arg_list:
| expr 
    { push($1); }
| caller_arg_list TK_COMMA expr
    { push($3); }
;

function: 
  TK_FUNCTION TK_ID TK_LPAR decl_args TK_RPAR decl_local_vars 
    { lbl($2); print_symbols(); emit_prelude(nb_local_vars); }
  TK_BEGIN stmt_list TK_END
    { emit_postlude(nb_arguments,nb_local_vars); }
;

function_list:
| function_list function
;



/* -------------------------------------------------------------- */
/*      5/ *Main production* (program=declarations+statements)    */
/* -------------------------------------------------------------- */

prog: function_list function
;
